"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.generatePersistedQueryManifest = void 0;
const cache_1 = require("@apollo/client/cache");
const core_1 = require("@apollo/client/core");
const persisted_query_lists_1 = require("@apollo/persisted-query-lists");
const graphql_tag_pluck_1 = require("@graphql-tools/graphql-tag-pluck");
const glob_1 = require("glob");
const graphql_1 = require("graphql");
const lodash_1 = require("lodash");
const node_crypto_1 = require("node:crypto");
const node_fs_1 = require("node:fs");
async function generatePersistedQueryManifest(options = {}) {
    const files = new Set(options.files ?? []);
    const globInclude = options.globInclude ?? [
        "src/**/*.{graphql,tsx,jsx,ts,js}",
    ];
    const globExclude = options.globExclude ?? [
        "**/*.spec.{js,jsx,ts,tsx}",
        "**/*.story.{js,jsx,ts,tsx}",
        "**/*.test.{js,jsx,ts,tsx}",
    ];
    for (const include of globInclude) {
        for (const f of await (0, glob_1.glob)(include, { ignore: globExclude ?? [] })) {
            files.add(f);
        }
    }
    const documents = [...files].flatMap((filePath) => {
        const content = (0, node_fs_1.readFileSync)(filePath, "utf-8");
        if (filePath.endsWith("graphql")) {
            return [{ ast: (0, graphql_1.parse)(content), filename: filePath }];
        }
        else {
            return (0, graphql_tag_pluck_1.gqlPluckFromCodeStringSync)(filePath, (0, node_fs_1.readFileSync)(filePath, "utf-8")).map((source) => ({ ast: (0, graphql_1.parse)(source.body), filename: source.name }));
        }
    });
    const seenFragmentNames = new Set();
    const operationsByName = new Map();
    for (const document of documents) {
        for (const definition of document.ast.definitions) {
            if (definition.kind === "FragmentDefinition") {
                const name = definition.name.value;
                if (seenFragmentNames.has(name)) {
                    throw new Error(`Duplicate fragment name: ${name}`);
                }
                seenFragmentNames.add(name);
            }
            else if (definition.kind === "OperationDefinition") {
                const name = definition.name?.value;
                if (!name) {
                    throw new Error("Anonymous operations not supported");
                }
                if (operationsByName.has(name)) {
                    throw new Error(`Duplicate operation name '${name}' in ${operationsByName.get(name)?.filename} and ${document.filename}`);
                }
                operationsByName.set(name, {
                    node: definition,
                    filename: document.filename,
                });
            }
        }
    }
    const fragments = (0, cache_1.createFragmentRegistry)(...documents.map(({ ast }) => ast));
    const manifestOperations = [];
    const client = new core_1.ApolloClient({
        cache: new core_1.InMemoryCache({
            fragments,
        }),
        link: new core_1.ApolloLink((operation) => {
            const body = (0, graphql_1.print)((0, persisted_query_lists_1.sortTopLevelDefinitions)(operation.query));
            manifestOperations.push({
                id: (0, node_crypto_1.createHash)("sha256").update(body).digest("hex"),
                name: operation.operationName,
                type: operation.query.definitions.find((d) => d.kind === "OperationDefinition").operation,
                body,
            });
            return core_1.Observable.of({ data: null });
        }),
    });
    for (const [_, operation] of (0, lodash_1.sortBy)([...operationsByName.entries()], lodash_1.first)) {
        const document = {
            kind: graphql_1.Kind.DOCUMENT,
            definitions: [operation.node],
        };
        await client.query({ query: document, fetchPolicy: "no-cache" });
    }
    return {
        format: "apollo-persisted-query-manifest",
        version: 1,
        operations: manifestOperations,
    };
}
exports.generatePersistedQueryManifest = generatePersistedQueryManifest;
//# sourceMappingURL=index.js.map